#########################################################################
#									#
# Script:	datechooser.tcl (namespace datechooser)			#
#									#
# Started:	23rd May, 2000.						#
#									#
# Completed:	25th May, 2000.						#
#									#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd		#
#									#
# Provides:	datechooser <widget path> [options]			#
#									#
#		The command/widget supports the following options,	#
#		(more to be added at later date).			#
#		-font		Used used for top labels.		#
#		-datefont	Font used for number labels.		#
#		-bg		Background for top labels.		#
#		-fg		Foreground for top labels.		#
#		-datebg		Background for number labels.		#
#		-datefg		Foreground for number labels.		#
#		-textvariable	Variable set when date is selected.	#
#		-highlightdate	Highlight the current date with 	#
#				different font / background.		#
#		-highlightfont	The font to use if highlighting the	#
#				current date.				#
#		-highlightbg	The background color if highlighting	#
#				the current date.			#
#		-setdates	Indicate a list of dates that should	#
#				be highlighted, defaulting to current	#
#				date if empty option.			#
#									#
#		Supported commands:					#
#									#
#		get		Return the last selected date.		#
#		set		Set the month to show, give date in	#
#				the format "1 month year".		#
#		config		Configure the widget options.		#
#		backward	Move the widget back a month.		#
#		forward		Move the widget forward a month.	#
#									#
# Purpose:	Providea a mega-widget allowing a particular date	#
#		to be chosen. Intended to be part of the crontool	#
#		and the capacity planning analyse GUI.			#
#									#
# Credits:	Bryan Oakley (oakley@channelpoint.com)			#
#		Shamelessly hacked his implementation of the		#
#		combobox widget to cobble together this widget!		#
#									#
# Limitations:	This is the initial version - it is neither pretty	#
#		nor fully functional - though it does the basics.	#
#									#
# Exit Codes:	0	Ok						#
#									#
# Version:	@(#)0.1.0 Original released version (SE)>		#
# 		@(#)0.2.0 Added highlight font and date options (SE)>	#
# 		@(#)0.3.0 Added work-around for tcl date bug (SE)>	#
# 		@(#)1.6 Support for multiple highlight dates (SE)>	#
#									#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
#									#
# Version:	1.6 (CVS)						#
# Date:		15th November, 2000.					#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd.		#
# Change:	Added support for the -setdates option, allowing	#
#		arbitary dates to be highlighted, not just current	#
#		date. Also includes several bug fixes, (see CHANGES.txt	#
#		document for details).					#
#									#
# Version:	0.4.0							#
# Date:		21st October, 2000.					#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd		#
# Change:	Use the "winfo rgb" command to darken the highlight	#
#		colour in an effort to make compatible with Windows.	#
#									#
# Version:	0.3.0							#
# Date:		19th October, 2000.					#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd		#
# Change:	Added a work around so that dates in October every	#
#		year work correctly. This works around a bug found in	#
#		tcl to at least version 8.3.				#
#									#
# Version:	0.2.0							#
# Date:		18th June, 2000.					#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd		#
# Change:	Added the highlightdate option and highlightfont	#
#		which can be used for highlighting the current date.	#
#									#
# Version:	0.1.0							#
# Date:		25th May, 2000.						#
# Author:	Simon Edwards, (C) Proprius Consulting Ltd		#
# Change:	Original released version.				#
#									#
# Version:								#
# Date:									#
# Author:								#
# Change:								#
#									#
#########################################################################

###############################################################################
#    datechooser.tcl - tcl module to provide datechooser mega-wdget           #
#    Copyright (C) 2000 Simon Edwards, Proprius Consulting Ltd                #
#                                                                             #
#    This program is free software; you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation; either version 2 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program; if not, write to the Free Software              #
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA  #
#                                                                             #
#    If you wish to contact the author the following details can be used;     #
#                                                                             #
#    simon.edwards@proprius.co.uk                                             #
###############################################################################

package require Tk 8.0
package provide datechooser 0.8

# Implements a date chooser widget...

namespace eval ::datechooser {
	namespace export datechooser

	variable widgetOptions
	variable widgetCommands
}

proc ::datechooser::darken {w col} {
	set x [winfo rgb $w $col]
	set x1 [expr round([lindex $x 0] * .95)]
	set x2 [expr round([lindex $x 1] * .95)]
	set x3 [expr round([lindex $x 2] * .95)]
	set ncol [format "%02x%02x%02x" $x1 $x2 $x3]
	return $ncol
}

proc ::datechooser::datechooser {w args} {
variable widgetOptions
variable widgetCommands

	# perform a one time initialization
		if {![info exists widgetOptions]} {
		Init
	}

	# We implement the new object as an object in its own
	# hidden namespace at present..

	eval Build $w $args

	return $w
}

proc ::datechooser::Init {} {
	variable widgetOptions
	variable widgetCommands

	array set widgetOptions [list \
		-background		{background	Background} 	\
		-bd			-borderwidth 			\
		-bg			-background 			\
		-borderwidth		{borderwidth	BorderWidth}	\
		-fg			-foreground 			\
		-font			{font		Font}		\
		-datefont		{datefont	DateFont}	\
		-datebg			{datebg		Datebg}		\
		-datefg			{datefg		Datefg}		\
		-dbg			-datebg				\
		-dfg			-datefg				\
		-foreground		{foreground	Foreground} 	\
		-setdates		{setdates	Setdates} 	\
		-textvariable		{textVariable	Variable}	\
		-highlightdate		{highlightdate	HighlightDate}	\
		-highlightfont		{highlightfont	HighlightFont}	\
		-highlightbg		{highlightbg	Highlightbg}	\
	]

	set widgetCommands [list \
		backward 	configure	forward		get 	\
		set							\
	]

	image create bitmap ::datechooser::image1 -data {
		#define test.bmp_width 10
		#define test.bmp_height 10
		static unsigned char test.bmp_bits[] = {
			   0x80,0x00,0xc0,0x00,0xe0,0x00,0xf0,0x00,0xf8,
			   0x00,0xf8,0x00,0xf0,0x00,0xe0,0x00,0xc0,0x00,
			   0x80,0x00;
		}
	}
	image create bitmap ::datechooser::image2 -data {
		#define test2_width 10
		#define test2_height 10
		static unsigned char test2_bits[] = {
		0x04,0x00,0x0c,0x00,0x1c,0x00,0x3c,0x00,0x7c,0x00,0x7c,0x00,
		0x3c,0x00,0x1c,0x00,0x0c,0x00,0x04,0x00;
		}
	}

	# Options in the Xresources would look like the following...
	# option add *Datechooser.foreground		black
	# option add *Datechooser.datefont		fixed
	# option add *Datechooser.highlightfont		fixed
	# option add *Datechooser.highlightdate		1
	# option add *Datechooser.datebg			green
	# option add *Datechooser.datefg			black
}

proc ::datechooser::Build {w args} {
variable widgetOptions
variable widgetCommands

	if {[winfo exists $w]} {
		error "window name \"$w\" already exists"
	}

	namespace eval ::datechooser::$w {

		variable ctime [clock seconds]
		variable cmonth [clock format $ctime -format "%B"]
		variable cyear [clock format $ctime -format "%Y"]

		variable widgets
		variable options
		variable chosen ""
		variable tvar ""

		set widgets(boo) boo
		set options(boo) boo

		unset widgets(boo)
		unset options(boo)
	}

	# Import variables from that proc here to allow
	# easy settings...

	upvar ::datechooser::${w}::widgets widgets
	upvar ::datechooser::${w}::options options
	upvar ::datechooser::${w}::cmonth cmonth
	upvar ::datechooser::${w}::cyear cyear

	set widgets(this)	[frame $w -class Datechooser -takefocus 0]
	set widgets(f1)		[frame $w.f1 -takefocus 0]
	set widgets(f2)		[frame $w.f2 -takefocus 0]
	set widgets(prev)	[button $w.f1.prev -image ::datechooser::image1 \
				-takefocus 0 \
					-command "::datechooser::prev_month $w"]
	set widgets(next)	[button $w.f1.next -image ::datechooser::image2 \
					-takefocus 0 \
					-command "::datechooser::next_month $w"]
	
	set widgets(label)	[label $w.f1.lab -width 14 -text "$cmonth $cyear" -takefocus 0]
	# Create the buttons for the individual dates...
	set days(0) "Sun"
	set days(1) "Mon"
	set days(2) "Tue"
	set days(3) "Wed"
	set days(4) "Thu"
	set days(5) "Fri"
	set days(6) "Sat"

	for {set x 0} {$x < 7} {incr x} {
		for {set y 0} {$y < 7} {incr y} {
			if { $y == 0 } {
				label $widgets(f2).b${y}_$x -text $days($x)
			} else {
				button $widgets(f2).b${y}_$x -width 2 -relief groove
			}
			grid $widgets(f2).b${y}_$x -in $widgets(f2) -row $y -column $x -sticky nsew
		}
	}

	# Fill in current buttons...

	::datechooser::update_buts $w

	# When we get around to having options, deal with them here

	# Store the proc to control the widget

	set widgets(frame) ::datechooser::${w}::${w}

	# Pack the widgets...

	pack $widgets(f1) -side top -fill x -expand yes
	pack $widgets(f2) -side bottom -fill both -expand yes
	pack $widgets(prev) -side left -fill none -expand yes
	pack $widgets(label) -side left -fill none -expand yes
	pack $widgets(next) -side right -fill none -expand yes

	# All options should have been loaded in the options array,
	# so deal with that stuff here...

	foreach name [array names widgetOptions] {
		if {[llength $widgetOptions($name)] == 1} continue
		set optName  [lindex $widgetOptions($name) 0]
		set optClass [lindex $widgetOptions($name) 1]
		set value [option get $w $optName $optClass]
		set options($name) $value
	}

	if {[info exists options(-value)] \
		&& [string length $options(-value)] == 0} {
		unset options(-value)

	}

	array set options $args

	rename ::$w $widgets(frame)

	# Set the class bindinsg for the widget
	SetClassBindings

	# Now create the proc to handle the widget...

	proc ::$w {command args} \
		"eval ::datechooser::WidgetProc $w \$command \$args"

	# Now do some configuration...

	if { [catch "::datechooser::Configure $widgets(this) [array get options]" error] } {
		catch {destroy $w}
		error $error
	}

	return ""
}

proc ::datechooser::SetClassBindings {} {

	bind Datechooser <Destroy> [list ::datechooser::DestroyHandler %W]
}

# Remove the name space / widgets associated with this widget...
proc ::datechooser::DestroyHandler {w} {

	if {[string compare [winfo class $w] "Combobox"] == 0} {
		upvar ::combobox::${w}::widgets  widgets
		upvar ::combobox::${w}::options  options

		# delete the namespace and the proc which represents
		# our widget
		namespace delete ::combobox::$w
		rename $w {}
	}  

	return ""
}

proc ::datechooser::Configure {w args} {
	variable widgetOptions

	upvar ::datechooser::${w}::widgets widgets
	upvar ::datechooser::${w}::options options
	upvar ::datechooser::${w}::tvar tvar
	upvar ::datechooser::${w}::cmonth cmonth
	upvar ::datechooser::${w}::cyear cyear

	# Return all results to the user
	if {[llength $args] == 0} {
		set results {}
		foreach opt [lsort [array names widgetOptions]] {
			if {[llength $widgetOptions($opt)] == 1} {
				set alias $widgetOptions($opt)
				set optName $widgetOptions($alias)
				lappend results [list $opt $optName]
			} else {
				set optName  [lindex $widgetOptions($opt) 0]
				set optClass [lindex $widgetOptions($opt) 1]
				set default [option get $w $optName $optClass]
				lappend results [list $opt $optName $optClass \
					$default $options($opt)]
			}
		}

		return $results
	}

	# Are we looking for config info for an option?
	if {[llength $args] == 1} {
		set opt [::datechooser::Canonize $w option [lindex $args 0]]

		set optName  [lindex $widgetOptions($opt) 0]
		set optClass [lindex $widgetOptions($opt) 1]
		set default [option get $w $optName $optClass]
		set results [list $opt $optName $optClass \
			$default $options($opt)]
		return $results
	}

	# if we have an odd number of values, bail.
	if {[expr {[llength $args]%2}] == 1} {
		# hmmm. An odd number of elements in args
		error "value for \"[lindex $args end]\" missing"
	}

	# deal with arg/value pairs, putting them in an array...
	foreach {name value} $args {
		set name [::datechooser::Canonize $w option $name]
		set opts($name) $value
	}

	foreach option [array names opts] {
		set newValue $opts($option)
		if {[info exists options($option)]} {
			set oldValue $options($option)
		}

		switch -- $option {
			-background {
				if { "$newValue" != "" } {
					$widgets(frame)	configure -background $newValue
					$widgets(f1)	configure -background $newValue
					$widgets(prev)	configure -background $newValue
					$widgets(next)	configure -background $newValue
					$widgets(label)	configure -background $newValue
				}
				set options($option) $newValue
			}
			-foreground {
				if { "$newValue" != "" } {
					$widgets(prev)	configure -foreground $newValue
					$widgets(next)	configure -foreground $newValue
					$widgets(label)	configure -foreground $newValue
					set options($option) $newValue
				}
			}
			-datebg {
				if { "$newValue" != "" } {
					$widgets(f2) configure -background $newValue
					for {set x 0} {$x < 7} {incr x} {
						$widgets(f2) config -bg $newValue
						for {set y 0} {$y < 7} {incr y} {
							$widgets(f2).b${y}_$x config -bg $newValue
						}
					}
				}
				set options($option) $newValue
			}
			-datefg {
				if { "$newValue" != "" } {
					for {set x 0} {$x < 7} {incr x} {
						# $widgets(f2) config -fg $newValue
						for {set y 0} {$y < 7} {incr y} {
							$widgets(f2).b${y}_$x config -fg $newValue
						}
					}
				}
				set options($option) $newValue
			}
			-setdates {
				set options($option) $newValue
				if { [info exists options(-highlightdate)] } {
					::datechooser::Configure $w -highlightdate $options(-highlightdate)
				}
			}
			-textvariable {
				set options($option) $newValue
				set tvar $newValue
			}
			-highlightdate {
				if { $newValue == "" } {
					set newValue 0
				}
				if { ! [info exists options(-highlightbg)] || \
						$options(-highlightbg) == "" } {
					set nv [lindex [$widgets(f2).b0_0 configure -background] 4]
					set nv [::datechooser::darken $widgets(label) $nv]
					set options(-highlightbg) "#$nv"
				} else {
					set nv $options(-highlightbg)
				}
				set options($option) $newValue
				::datechooser::update_buts $w
				return
			}
			-font {
				if { $newValue == "" } {
					set newValue [font create -slant roman -weight normal -family Helvetica -size 12]
				}
				$widgets(prev)	configure -font $newValue
				$widgets(next)	configure -font $newValue
				$widgets(label)	configure -font $newValue
				set options($option) $newValue
			}
			-datefont {
				if { $newValue == "" } {
					set newValue [font create -slant roman -weight normal -family Helvetica -size 10]
				}
				for {set x 0} {$x < 7} {incr x} {
					for {set y 0} {$y < 7} {incr y} {
						$widgets(f2).b${y}_$x config -font $newValue
					}
				}
				set options($option) $newValue
			}
			-highlightfont {
				if { $newValue == "" } {
					set newValue [font create -slant roman -weight bold -family Helvetica -size 10]
				}
				set options(-highlightfont) $newValue
				::datechooser::Configure $w -highlightdate $options(-highlightdate)
			}
			-highlightbg {
				if { $newValue == "" } {
					set newValue [lindex [$widgets(f2).b0_0 configure -background] 4]
				}
				set options($option) $newValue
				::datechooser::Configure $w -highlightdate $options(-highlightdate)
			}
		}
	}
}

proc ::datechooser::Canonize {w object opt} {
	variable	widgetOptions
	variable	widgetCommands

	switch $object {
		command {
			if {[lsearch -exact $widgetCommands $opt] >= 0} {
				return $opt
			}
			# stuff command names in temporary array 
			set list $widgetCommands
			foreach element $list {
				set tmp($element) ""
			}
			set matches [array names tmp ${opt}*]
		}

		option {
			if {[info exists widgetOptions($opt)] \
				&& [llength $widgetOptions($opt)] == 2} {
				return $opt
			}
			set list [array names widgetOptions]
			set matches [array names widgetOptions ${opt}*]
		}
	}

	if {[llength $matches] == 0} {
		set choices [HumanizeList $list]
		error "unknown $object \"$opt\"; must be one of $choices"
	} elseif {[llength $matches] == 1} {
		set opt [lindex $matches 0]

		# deal with option aliases
		switch $object {
			option {
				set opt [lindex $matches 0]
				if {[llength $widgetOptions($opt)] == 1} {
					set opt $widgetOptions($opt)
				}
			}
		}

		return $opt
	} else {
		set choices [HumanizeList $list]
		error "ambiguous $object \"$opt\"; must be one of $choices"
	}
}

proc ::datechooser::HumanizeList {list} {

	if {[llength $list] == 1} {
		return [lindex $list 0]
	} else {
		set list [lsort $list]
		set secondToLast [expr {[llength $list] -2}]
		set most [lrange $list 0 $secondToLast]
		set last [lindex $list end]

		return "[join $most {, }] or $last"
	}
}

# Handle any sub-widget stuff and get errors if required, so it
# appears to come from our "widget"

proc ::datechooser::DoInternalWidgetCommand {w subwidget command args} {
    upvar ::datechooser::${w}::widgets widgets
    upvar ::datechooser::${w}::options options

    set subcommand $command
    set command [concat $widgets($subwidget) $command $args]
    if {[catch $command result]} {
        # replace the subwidget name with the megawidget name
        regsub $widgets($subwidget) $result $widgets(this) result

        # replace specific instances of the subwidget command
        # with out megawidget command
        switch $subwidget,$subcommand {
            listbox,index  {regsub "index"  $result "list index"  result}
            listbox,insert {regsub "insert" $result "list insert" result}
            listbox,delete {regsub "delete" $result "list delete" result}
            listbox,get    {regsub "get"    $result "list get"    result}
            listbox,size   {regsub "size"   $result "list size"   result}
        }
        error $result

    } else {
        return $result
    }
}

# This is the actual command that gets called when the user attempts
# to use the widget object name as a procedure.

proc ::datechooser::WidgetProc {w command args} {
upvar ::datechooser::${w}::widgets widgets
upvar ::datechooser::${w}::options options
upvar ::datechooser::${w}::chosen chosen
upvar ::datechooser::${w}::cmonth cmonth
upvar ::datechooser::${w}::cyear cyear

	set command [::datechooser::Canonize $w command $command]

	set doWidgetCommand \
		[list ::datechooser::DoInternalWidgetCommand $widgets(this)]

	set result ""

	switch $command {
		configure {
			set result [eval ::datechooser::Configure {$w} $args]
		}
		set {
			# Set the date to look at ...
			set xx [clock scan "[lindex $args 0]"]
			if { [catch {set xx [clock scan "[lindex $args 0]"]}] } {
				error "Invalid date specified for \"set\" command."
			} else {
				set cmonth [clock format $xx -format "%B"]
				set cyear [clock format $xx -format "%Y"]
				# Now change the dates shown...
				update_buts $w
			}
		}
		backward {
			::datechooser::prev_month $w
		}
		forward {
			::datechooser::next_month $w
		}
		get {
			return "$chosen"
		}
		default {
			error "bad option \"$command\""
		}
	}
}

# This is called when the month changes, or the widget is first
# created...

proc ::datechooser::update_buts {w} {
upvar ::datechooser::${w}::widgets widgets
upvar ::datechooser::${w}::cmonth cmonth
upvar ::datechooser::${w}::cyear cyear
upvar ::datechooser::${w}::options options

	if { ! [info exists options(-datefont)] } {
		set options(-datefont) [font create -slant roman -weight normal -family Helvetica -size 12]
	}
	if { ! [info exists options(-highlightdate)] } {
		set options(-highlightdate) 0
	}
	# Firstly remove the dates from the first and last weeks...
	# 0.8 - and also set the bg to the default colour...
	if { ! [info exists options(-datebg)] || "$options(-datebg)" == "" } {
		set nv [lindex [$widgets(f2).b0_0 configure -background] 4]
		set options(-datebg) $nv
	}

	for {set x 0} {$x < 7} {incr x} {
		foreach y "1 5 6" { 
			$widgets(f2).b${y}_$x config -text "" -state disabled \
				-bg $options(-datebg)
		}
	}

	set normalval [lindex [$widgets(f2).b0_0 configure -background] 4]
	if { ! [info exists options(-highlightbg)] } {
		set nv [lindex [$widgets(f2).b0_0 configure -background] 4]
	} else {
		set nv $options(-highlightbg)
	}
	# Date the month details...
	set secs [clock scan "1 $cmonth $cyear"]
	set osecs $secs
	set mdays "Sun Mon Tue Wed Thu Fri Sat"
	set y 1
	set dom 1
	set c_year [clock format $secs -format "%Y"]
	if { [info exists options(-setdates)] && "$options(-setdates)" != "" } {
		set cdates $options(-setdates)
	} else {
		set cdates [clock format [clock seconds] -format "%d %B %Y"]
		if { [string index $cdates 0] == "0" } {
			set cdates [string range $cdates 1 end]
		}
		set cdates "{$cdates}"
	}

	# Following line for October date problem in tcl 8.3
	set prev_x -1

	while {1} {
		if { $prev_x == -1 } {
			set c_day [clock format $secs -format "%a"]
		}

		set c_month [clock format $secs -format "%B"]
		if { "$c_month" != "$cmonth" } break

		if { $prev_x == -1 } {
			set x [lsearch -exact $mdays $c_day]
		} else {
			incr x 1
			if { $x == 7 } {
				set x 0
			}
		}
		set prev_x $x
		set _cdate "$dom $cmonth $cyear"

		if { $options(-highlightdate) == 1 && \
			[lsearch -exact $cdates $_cdate] > -1 } {
			$widgets(f2).b${y}_$x config -text "$dom" -state normal \
				-command "::datechooser::but_hit $w \"$dom $c_month $c_year\"" \
				-font $options(-highlightfont) -bg $options(-highlightbg)
		} else {
			$widgets(f2).b${y}_$x config -text "$dom" -state normal \
				-command "::datechooser::but_hit $w \"$dom $c_month $c_year\"" \
				-font $options(-datefont) -bg $normalval
		}

		incr dom
		if { $x == 6 } { incr y }
		incr secs 86400

		# Added due to October bug in tcl 8.3
		if { $dom > 31 } break
	}
	$widgets(label) config -text "[clock format $osecs -format "%B %Y"]"
}


# Called to update the chosen and textvariable values as required...

proc ::datechooser::but_hit {w d} {
upvar ::datechooser::${w}::chosen chosen
upvar ::datechooser::${w}::tvar tvar

	set chosen "$d"
	if { "$tvar" != "" } {
		global $tvar
		set $tvar $d
	}
}

proc ::datechooser::prev_month w {
upvar ::datechooser::${w}::cmonth cmonth
upvar ::datechooser::${w}::cyear cyear

	set mlist "January February March April May June July August September October November December"
	set cmon [lsearch -exact $mlist $cmonth]
	if { $cmon == 0 } {
		set cmon 11
		incr cyear -1
	} else {
		incr cmon -1
	}
	set cmonth [lindex $mlist $cmon]
	update_buts $w
}

proc ::datechooser::next_month w {
upvar ::datechooser::${w}::cmonth cmonth
upvar ::datechooser::${w}::cyear cyear

	set mlist "January February March April May June July August September October November December"
	set cmon [lsearch -exact $mlist $cmonth]
	if { $cmon == 11 } {
		set cmon 0
		incr cyear
	} else {
		incr cmon
	}
	set cmonth [lindex $mlist $cmon]
	update_buts $w
}

